#define SETS_C
#include "sets.h"

/* this function allocates memory for set elements if none is available
   in the SetElementPool */
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
SetElement     *SetElementCalloc()
{
   SetElement     *setelement;

/* check whether or not there is memory available in SetElementPool */
   if (SetElementPool->First != NULL)
   {
      setelement = SetElementPool->First;
      SetElementPool->First = setelement->Next;
      (SetElementPool->n)--;
   }                            /* end if */
   else
   {
      setelement = (SetElement *) calloc(1, sizeof(SetElement));
   }                            /* end else */

   return setelement;

}                               /* end *SetElementCalloc() */


/* this function takes memory to be freed and places that
   memory in the set SetElementPool to be reused later */
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void            SetElementFree(SetElement *setelement)
{
   setelement->Next = SetElementPool->First;
   SetElementPool->First = setelement;
   (SetElementPool->n)++;
}                               /* end SetElementFree() */


/* this function allocates memory for set-of-set elements if none
   is available in the SetOfSetsElementPool */
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
SoSElement     *SetOfSetsElementCalloc()
{
   SoSElement     *soselement;

/* check whether or not there is memory available in SetOfSetsElementPool */
   if (SetOfSetsElementPool->First != NULL)
   {
      soselement = SetOfSetsElementPool->First;
      SetOfSetsElementPool->First = soselement->Next;
      (SetOfSetsElementPool->n)--;
   }                            /* end if */
   else
   {
      soselement = (SoSElement *) calloc(1, sizeof(SoSElement));
   }                            /* end else */

   return soselement;

}                               /* end *SetOfSetsElementCalloc() */


/* this function takes memory to be freed and places that
   memory in the set-of-sets SetOfSetsElementPoolto be reused later */
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void            SetOfSetsElementFree(SoSElement *soselement)
{
   soselement->Next = SetOfSetsElementPool->First;
   SetOfSetsElementPool->First = soselement;
   (SetOfSetsElementPool->n)++;
}                               /* end SetOfSetsElementFree */


/* this function initializes a set */
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
Set            *InitSet(Set *set, char *string)
{
   SetElement     *pres, *next;

/* determine whether or not the memory for the set has been allocated */
   if (set == NULL)
   {
      set = (Set *) calloc(1, sizeof(Set));
      if (set == NULL)
      {
         printf("can`t allocate memory for set \n");
         exit(1);
      }                         /* end if */
   }                            /* if memory for set has not yet been
                                 * allocated */

/* determine whether or not there are any elements in the set
   if so delete and free the memory for those elements */
   if (set->First != NULL)
   {
      pres = set->First;
      while (pres != NULL)
      {
         next = pres->Next;
         SetElementFree(pres);
         pres = next;
      }                         /* end while */
   }                            /* end if */
   set->n = 0;
   set->nOpsBlockGS = 0;
   set->nOpsBorderGS = 0;
   set->nOpsTotalGS = 0;
   set->SuperCluster = NO;
   set->Name = string;
   set->First = NULL;

   return set;

}                               /* end InitSet() */


/* this function deletes a set and frees the memory */
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void            DeleteSet(Set *set)
{
   SetElement     *pres, *next;

/* determine whether or not the memory for the set has been allocated */
   if (set != NULL)
   {
/* determine whether or not there are any elements in the set
   if so delete and free the memory for those elements */
      if (set->First != NULL)
      {
         pres = set->First;
         while (pres != NULL)
         {
            next = pres->Next;
            SetElementFree(pres);
            pres = next;
         }                      /* end while */
      }                         /* end if */
      set->n = 0;
      set->Name = "";
      set->First = NULL;
   }                            /* if memory for set has been allocated */

}                               /* end DeleteSet() */


/* this function initializes a set of sets */
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
SoSets         *InitSetOfSets(SoSets *sosets, char *string)
{
   SoSElement     *pres, *next;

/*
printf("InitSetOfSets() - string=|%s| \n",string);
fflush(stdout);
*/
/* determine whether or not the memory for the set of sets has been allocated */
   if (sosets == NULL)
   {
      sosets = (SoSets *) calloc(1, sizeof(SoSets));
      if (sosets == NULL)
      {
         printf("can`t allocate memory for sosets \n");
         exit(1);
      }                         /* end if */
   }                            /* if memory for sosetss has not yet been
                                 * allocated */

/* determine whether or not there are any elements in the set of sets
   if so delete and free the memory for those elements */
   if (sosets->First != NULL)
   {
      pres = sosets->First;
      while (pres != NULL)
      {
         next = pres->Next;
         SetOfSetsElementFree(pres);
         pres = next;
      }                         /* end while */
   }                            /* end if */

   sosets->n = 0;
   sosets->Name = string;
   sosets->First = NULL;

   return sosets;

}                               /* end InitSetOfSets() */


/* this function takes the union of set1 and set2 and places the
   result in set1  - the operation does not destroy set2 */
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void            UnionOfSets(Set *set1, Set *set2)
{
   int             element;
   SetElement     *pres2;

   pres2 = set2->First;
   while (pres2 != NULL)
   {
      element = pres2->Element;
      AddElementToSet(set1, element);
      pres2 = pres2->Next;
   }                            /* end while */
}                               /* end UnionOfSets() */


/* this function takes the difference of set1 and set2 and places
   the result in set1 */
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void            DifferenceOfSets(Set *set1, Set *set2)
{
   int             element;
   SetElement     *pres2;

   pres2 = set2->First;
   while (pres2 != NULL)
   {
      element = pres2->Element;
      DeleteElementFromSet(set1, element);
      pres2 = pres2->Next;
   }                            /* end while */
}                               /* end DifferenceOfSets() */


/* this function adds an integer-type element to the set
   the element is placed at the head of the set aka cdr */
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void            AddElementToSet(Set *set, int element)
{
   SetElement     *first, *next;

   if (ElementOfSet(set, element) == NO)
   {
      next = set->First;
      first = SetElementCalloc();
      if (first == NULL)
      {
         printf("can't allocate memory for first in AddElementToSet()\n");
         exit(1);
      }                         /* end if */
      first->Element = element;
      first->Next = next;
      set->First = first;
      (set->n)++;
   }                            /* end if ElementOfSet() == NO */

}                               /* AddElementToSet() */


/* this function removes an element from the set */
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void            DeleteElementFromSet(Set *set, int element)
{
   SetElement     *prev, *pres;

   if (set->First != NULL)
   {
      pres = set->First;
/* is the element to be deleted located in the first element location? */
      if (pres->Element == element)
      {
         set->First = pres->Next;
         SetElementFree(pres);
         (set->n)--;
      }                         /* end if */
/* is the element to be deleted located in any remaining location? */
      else
      {
         prev = set->First;
         pres = prev->Next;
         while (pres != NULL)
         {
            if (pres->Element == element)
            {
               prev->Next = pres->Next;
               SetElementFree(pres);
               (set->n)--;
               break;
            }                   /* end if */
            prev = pres;
            pres = pres->Next;
         }                      /* end while */
      }                         /* end else */

   }                            /* end if set->First != NULL */
}                               /* end DeleteElementFromSet() */


/* this function extracts the minimum value in the set */
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
int             MinValueInSet(Set *set)
{
   int             minvalue;
   int             elementof;
   SetElement     *pres;

   if ((set->n) > 0)
   {
      minvalue = INFI;
      pres = set->First;
      while (pres != NULL)
      {
         if (pres->Element < minvalue)
         {
            minvalue = pres->Element;
         }                      /* end if */
         pres = pres->Next;
      }                         /* end while */
      DeleteElementFromSet(set, minvalue);
   }                            /* end if */
   else
   {
      minvalue = 0;
   }                            /* end else */
   return minvalue;

}                               /* end MinValueInSet() */

/* this function determines whether or not element is an element of the set */
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
int             ElementOfSet(Set *set, int element)
{
   int             elementof;
   SetElement     *pres;

   elementof = NO;
   pres = set->First;
   while (pres != NULL)
   {
      if (pres->Element == element)
      {
         elementof = YES;
         break;
      }                         /* end if */
      pres = pres->Next;
   }                            /* end while */

   return elementof;
}                               /* end ElementOfSet() */


/* this function prints a set */
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void            PrintSet(Set *set)
{
   int             i;
   SetElement     *pres;

   i = 0;

   printf("%s={", set->Name);
   fflush(stdout);
   pres = set->First;
   while (pres != NULL)
   {
      printf("%d,", pres->Element);
      i++;
      if ((i % 20) == 0)
         printf("\n ");
      pres = pres->Next;
   }                            /* end while */
   printf("}\n");
   printf("|%s| = %d \n", set->Name, set->n);

}                               /* end PrintSet() */


/* this function places a set into a set of sets for storage
   the set is added to the head of the set of sets aka cdr */
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void            AddSetToSetOfSets(SoSets *sosets, Set *set)
{
   SoSElement     *first, *next;

   next = sosets->First;
   first = SetOfSetsElementCalloc();
   if (first == NULL)
   {
      printf("can't allocate memory for first in AddSetToSetOfSets()\n");
      exit(1);
   }                            /* end if */
   first->set = set;
   first->Next = next;
   sosets->First = first;
   (sosets->n)++;
}                               /* end AddSetToSetOfSets() */


/* this function prints a set of sets */
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void            PrintSetOfSets(SoSets *sosets)
{
   int             i;
   SoSElement     *pres;

   i = 0;

   printf("%s={\n", sosets->Name);
   pres = sosets->First;
   while (pres != NULL)
   {
      printf("set %d of %d \n", ++i, sosets->n);
      PrintSet(pres->set);
      pres = pres->Next;
   }                            /* end while */
   printf("}\n");
   printf("|%s| = %d \n", sosets->Name, sosets->n);

}                               /* PrintSetOfSets() */

